<?php

/**
 * @file
 *  Drush command for starting and stopping a webserver on localhost for
 *  development purposes.
 */

define('DRUSH_INSTAWEB_TEMPLATES', 'templates');

/**
 * Implementation of hook_drush_command
 */
function instaweb_drush_command() {
  $items['instaweb'] = array(
    'description' => "Run a development webserver for this drupal instance",
    'options' => array(
      'bind' => 'Bind the testserver to the specified hostname or ip address.',
      'port' => 'Use specified TCP port (>= 1000)',
      'server' => 'Server to use. By default the first one detected on your system.',
      'rundir' => 'Path to directory where configuration is stored. By default a temporary directory is generated.',
      'tpldir' =>  'Path to config templates directory. Default: ' .
      DRUSH_INSTAWEB_TEMPLATES,
    ),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
  );

  return $items;
}

/**
 * Implementation of drush_hook_COMMAND_validate
 */
function drush_instaweb_validate() {
  $servers = array_merge(_instaweb_load_servers(), drush_get_option('servers', array()));

  if (empty($servers)) {
    return drush_set_error('NO_SERVER_FOUND',
      dt('No suitable server software found on your system. Either install apache or lighttpd.'));
  }

  $name = drush_get_option('server', key($servers));

  if (empty($servers[$name])) {
    return drush_set_error('NO_SERVER_CONFIG',
      dt('No suitable server config was found with the given server-name'));
  }

  $server = $servers[$name];

  // Add server name
  if (empty($server['name'])) {
    $server['name'] = $name;
  }

  // Check whether the server binary is set and is executable.
  if (empty($server['cmd'])) {
    return drush_set_error('SPECIFIED_SERVER_NO_CMD',
      dt('The server-record is incomplete. The path to the server binary is necessary'));
  }
  if (!is_executable($server['cmd'])) {
    return drush_set_error('SPECIFIED_SERVER_NOT_EXECUTABLE',
      dt('The specified path to the server command is invalid. Either the binary is missing or it is not executable'));
  }

  $bind = drush_get_option('bind');
  if ($bind) {
    $server['bind'] = $bind;
  }

  $port = drush_get_option('port');
  if ($port) {
    $server['port'] = $port;
  }

  // Validate port
  $port = intval($server['port']);
  if ($port < 1000 || $port > 0xFFFF) {
    return drush_set_error('PORT_OUT_OF_RANGE',
      dt('Please specify a port in the unprivileged range between 1000 and 65535'));
  }

  // Add root directory
  if (empty($server['root'])) {
    drush_bootstrap(DRUSH_BOOTSTRAP_DRUPAL_ROOT);
    $server['root'] = drush_get_context('DRUSH_DRUPAL_ROOT');
  }

  // Config directory
  $rundir = drush_get_option('rundir');
  if ($rundir) {
    $server['rundir'] = $rundir;
  }

  // Prepare temporary rundir if necessary
  if (empty($server['rundir'])) {
    $server['rundir'] = drush_tempdir();
  }

  // Substitute server name with server record.
  drush_set_option('server', $server);
}

/**
 * Implementation of drush_hook_COMMAND
 */
function drush_instaweb() {
  $server = drush_get_option('server');
  $rundir = $server['rundir'];

  // Write missing template files
  foreach ($server['templates'] as $template) {
    $tplpath = $rundir . '/' . $template;
    if (!is_file($tplpath)) {
      file_put_contents($tplpath, _instaweb_render_template($template, $server));
    }
  }

  // Prepare command line
  $rundir_patterns = array();
  $rundir_replacements = array();
  foreach (scandir($rundir) as $filename) {
    if ($filename[0] == '.') {
      continue;
    }

    $rundir_patterns[] = '!' . $filename;
    $rundir_replacements[] = drush_escapeshellarg($rundir . '/' . $filename);
  }

  $command = $server['cmd'] . ' ' . str_replace($rundir_patterns, $rundir_replacements, $server['args']);

  // Start server
  drush_log(dt("Starting !name on port !bind:!port", array(
    "!name" => $server['name'], "!bind" => $server['bind'], "!port" => $server['port'])), 'notice');
  drush_log(dt("Press ctl-C to exit"), 'notice');

  $origmask = _instaweb_signals_ignore();
  drush_shell_exec_interactive($command);
  _instaweb_signals_restore($origmask);
}

/**
 * Let the subprocess handle SIGINT and ignore SIGCHLD and SIGTERM as long as a
 * subprocess is running.
 */
function _instaweb_signals_ignore() {
  if (function_exists("pcntl_signal")) {
    // Ignore SIGINT / SIGTERM: Pass them to child process
    pcntl_signal(SIGCHLD, SIG_IGN);
    pcntl_signal(SIGINT, SIG_IGN);
    pcntl_signal(SIGTERM, SIG_IGN);

    // Block SIGWINCH (Window size change)
    pcntl_sigprocmask(SIG_BLOCK, array(SIGWINCH), $procmask);
    return $procmask;
  }
}

/**
 * Restore default signal handlers after a subprocess terminated.
 */
function _instaweb_signals_restore($procmask) {
  if (function_exists("pcntl_signal")) {
    pcntl_signal(SIGCHLD, SIG_DFL);
    pcntl_signal(SIGINT, SIG_DFL);
    pcntl_signal(SIGTERM, SIG_DFL);
    pcntl_signal(SIGWINCH, SIG_DFL);
    pcntl_sigprocmask(SIG_SETMASK, $procmask);
  }
}

/**
 * Default servers
 */
function _instaweb_default_servers() {
  return array(
    'apache' => array(
      'discover' => array(
        array(
          'target' => 'cmd',
          'conditions' => array('is_file', 'is_executable'),
          'paths' => array(
            '/sbin/apache2',
            '/sbin/httpd',
          ),
        ),
        array(
          'target' => 'libdir',
          'conditions' => array('is_dir'),
          'paths' => array(
            '/lib/apache2/modules',
            '/libexec/apache2',
            '/libexec/apache',
          ),
        ),
      ),
      'args' => '-DNO_DETACH -DFOREGROUND -f !apache2.conf',
      'templates' => array(
        'apache2.conf',
      ),
      'bind' => 'localhost',
      'port' => '3000',
    ),

    'lighty' => array(
      'discover' => array(
        array(
          'target' => 'cmd',
          'conditions' => array('is_file', 'is_executable'),
          'paths' => array(
            '/sbin/lighttpd',
          ),
        ),
      ),
      'args' => '-D -f !lighttpd.conf',
      'templates' => array(
        'lighttpd.conf',
        'magnet.lua',
      ),
      'bind' => 'localhost',
      'port' => '3000',
    ),
  );
}

/**
 * Return an array of configured servers
 */
function _instaweb_load_servers() {
  $servers = _instaweb_default_servers();

  $result = array();
  foreach ($servers as $name => $server) {
    $msg = _instaweb_server_pre_discover_validate($server);
    if ($msg !== NULL) {
      drush_log($msg, 'error');
      continue;
    }

    $result[$name] =  _instaweb_discover_server($server);
  }

  return $result;
}

/**
 * Validate a server record before server discovery.
 */
function _instaweb_server_pre_discover_validate($server) {
  if (isset($server['discover'])) {
    if (!is_array($server['discover'])) {
      return dt('Server discover record must be an array');
    }

    foreach ($server['discover'] as $discover) {
      // Ensure that there is at least one path we can check
      if (empty($discover['paths']) || !is_array($discover['paths'])) {
        return dt('At least one path is required in the discovery paths array');
      }

      // Ensure that there is at least one condition
      if (empty($discover['conditions']) || !is_array($discover['paths'])) {
        return dt('At least one condition is required in the discovery conditions array');
      }

      // Ensure that condition array 
      if ($discover['conditions'] != array_filter($discover['conditions'], 'is_callable')) {
        return dt('Server discovery conditions must be an array with functions like is_file, is_dir or is_executable');
      }
    }
  }
}

/**
 * Resolve paths to server binaries
 */
function _instaweb_discover_server($server) {
  $prefixes = array(
    '/usr',
    '/usr/local',
    '/opt/local',
    '/sw',
  );

  if (!isset($server['discover'])) {
    return;
  }

  // Loop through prefixes and resolve discovery records.
  foreach ($prefixes as $pfx) {
    foreach ($server['discover'] as $discover) {
      $target = $discover['target'];

      // Skip discovery if the target path is already set in the server record.
      if (!empty($server[$target])) {
        continue;
      }

      foreach ($discover['paths'] as $path) {
        $path = $pfx . $path;

        $veto = FALSE;
        foreach ($discover['conditions'] as $func) {
          if (!$func($path)) {
            $veto = TRUE;
            break;
          }
        }

        if (!$veto) {
          $server[$target] = $path;
          break;
        }
      }
    }
  }

  unset($server['discover']);
  return $server;
}

/**
 * Render given template_file and return result. This function is a ripoff of
 * theme_render_template.
 */
function _instaweb_render_template($template, $variables) {
  $variables['template'] = $template;
  $variables['directory'] = drush_get_option('tpldir', DRUSH_INSTAWEB_TEMPLATES);

  // Extract the variables to a local namespace.
  extract($variables, EXTR_SKIP);               
  unset($variables);

  // Evaluate template file and return the content.
  ob_start();
  require $directory . '/' . $template . '.tpl.php';
  return ob_get_clean();
}
